@@ -26,9 +26,10 @@
 
 #include "php.h"
 #include "php_ini.h"
-#include "ext/standard/info.h"
-#include "zend_smart_str.h"
 #include "SAPI.h"
+#include "ext/standard/info.h"
+#include "Zend/zend_smart_str.h"
+#include "Zend/zend_exceptions.h"
 
 #include "php_yac.h"
 #include "storage/yac_storage.h"
@@ -41,8 +42,19 @@
 
 zend_class_entry *yac_class_ce;
 
+static zend_object_handlers yac_obj_handlers;
+
+typedef struct {
+	unsigned char prefix[YAC_STORAGE_MAX_KEY_LEN];
+	uint16_t prefix_len;
+	zend_object std;
+} yac_object;
+
 ZEND_DECLARE_MODULE_GLOBALS(yac);
 
+static yac_serializer_t yac_serializer;
+static yac_unserializer_t yac_unserializer;
+
 /** {{{ ARG_INFO
  */
 ZEND_BEGIN_ARG_INFO_EX(arginfo_yac_constructor, 0, 0, 0)
@@ -55,19 +67,10 @@ ZEND_BEGIN_ARG_INFO_EX(arginfo_yac_add, 0, 0, 1)
 	ZEND_ARG_INFO(0, ttl)
 ZEND_END_ARG_INFO()
 
-ZEND_BEGIN_ARG_INFO_EX(arginfo_yac_setter, 0, 0, 2)
-	ZEND_ARG_INFO(0, key)
-	ZEND_ARG_INFO(0, value)
-ZEND_END_ARG_INFO()
-
 ZEND_BEGIN_ARG_INFO_EX(arginfo_yac_get, 0, 0, 1)
 	ZEND_ARG_INFO(0, keys)
 ZEND_END_ARG_INFO()
 
-ZEND_BEGIN_ARG_INFO_EX(arginfo_yac_getter, 0, 0, 1)
-	ZEND_ARG_INFO(0, key)
-ZEND_END_ARG_INFO()
-
 ZEND_BEGIN_ARG_INFO_EX(arginfo_yac_delete, 0, 0, 1)
 	ZEND_ARG_INFO(0, keys)
 	ZEND_ARG_INFO(0, ttl)
@@ -113,38 +116,62 @@ PHP_INI_BEGIN()
     STD_PHP_INI_ENTRY("yac.values_memory_size", "64M", PHP_INI_SYSTEM, OnChangeValsMemoryLimit, v_msize, zend_yac_globals, yac_globals)
     STD_PHP_INI_ENTRY("yac.compress_threshold", "-1", PHP_INI_SYSTEM, OnChangeCompressThreshold, compress_threshold, zend_yac_globals, yac_globals)
     STD_PHP_INI_ENTRY("yac.enable_cli", "0", PHP_INI_SYSTEM, OnUpdateBool, enable_cli, zend_yac_globals, yac_globals)
+    STD_PHP_INI_ENTRY("yac.serializer", "php", PHP_INI_SYSTEM, OnUpdateString, serializer, zend_yac_globals, yac_globals)
 PHP_INI_END()
 /* }}} */
 
-static int yac_add_impl(zend_string *prefix, zend_string *key, zval *value, int ttl, int add) /* {{{ */ {
+#define Z_YACOBJ_P(zv)   (php_yac_fetch_object(Z_OBJ_P(zv)))
+static inline yac_object *php_yac_fetch_object(zend_object *obj) /* {{{ */ {
+	return (yac_object *)((char*)(obj) - XtOffsetOf(yac_object, std));
+}
+/* }}} */
+
+static const char *yac_assemble_key(yac_object *yac, zend_string *name, size_t *len) /* {{{ */ {
+	const char *key;
+
+	if ((ZSTR_LEN(name) + yac->prefix_len) > YAC_STORAGE_MAX_KEY_LEN) {
+		php_error_docref(NULL, E_WARNING,
+				"Key '%.*s%s' exceed max key length '%d' bytes",
+				yac->prefix_len, yac->prefix, ZSTR_VAL(name), YAC_STORAGE_MAX_KEY_LEN);
+		return NULL;
+	}
+
+	if (yac->prefix_len) {
+		memcpy(yac->prefix + yac->prefix_len, ZSTR_VAL(name), ZSTR_LEN(name));
+		key = (const char*)yac->prefix;
+		*len = yac->prefix_len + ZSTR_LEN(name);
+	} else {
+		key = ZSTR_VAL(name);
+		*len = ZSTR_LEN(name);
+	}
+
+	return key;
+}
+/* }}} */
+
+static int yac_add_impl(yac_object *yac, zend_string *name, zval *value, int ttl, int add) /* {{{ */ {
 	int ret = 0, flag = Z_TYPE_P(value);
 	char *msg;
 	time_t tv;
-	zend_string *prefix_key;
+	const char *key;
+	size_t key_len;
 
-	if ((ZSTR_LEN(key) + prefix->len) > YAC_STORAGE_MAX_KEY_LEN) {
-		php_error_docref(NULL, E_WARNING, "Key%s can not be longer than %d bytes",
-				prefix->len? "(include prefix)" : "", YAC_STORAGE_MAX_KEY_LEN);
+	if ((key = yac_assemble_key(yac, name, &key_len)) == NULL) {
 		return ret;
 	}
 
-	if (prefix->len) {
-		prefix_key = strpprintf(YAC_STORAGE_MAX_KEY_LEN, "%s%s", ZSTR_VAL(prefix), ZSTR_VAL(key));
-		key = prefix_key;
-	}
-
 	tv = time(NULL);
 	switch (Z_TYPE_P(value)) {
 		case IS_NULL:
 		case IS_TRUE:
 		case IS_FALSE:
-			ret = yac_storage_update(ZSTR_VAL(key), ZSTR_LEN(key), (char *)&flag, sizeof(int), flag, ttl, add, tv);
+			ret = yac_storage_update(key, key_len, (char *)&flag, sizeof(int), flag, ttl, add, tv);
 			break;
 		case IS_LONG:
-			ret = yac_storage_update(ZSTR_VAL(key), ZSTR_LEN(key), (char *)&Z_LVAL_P(value), sizeof(long), flag, ttl, add, tv);
+			ret = yac_storage_update(key, key_len, (char *)&Z_LVAL_P(value), sizeof(long), flag, ttl, add, tv);
 			break;
 		case IS_DOUBLE:
-			ret = yac_storage_update(ZSTR_VAL(key), ZSTR_LEN(key), (char *)&Z_DVAL_P(value), sizeof(double), flag, ttl, add, tv);
+			ret = yac_storage_update(key, key_len, (char *)&Z_DVAL_P(value), sizeof(double), flag, ttl, add, tv);
 			break;
 		case IS_STRING:
 #ifdef IS_CONSTANT
@@ -154,13 +181,10 @@ static int yac_add_impl(zend_string *prefix, zend_string *key, zval *value, int
 				if (Z_STRLEN_P(value) > YAC_G(compress_threshold) || Z_STRLEN_P(value) > YAC_STORAGE_MAX_ENTRY_LEN) {
 					int compressed_len;
 					char *compressed;
-				   
+
 					/* if longer than this, then we can not stored the length in flag */
 					if (Z_STRLEN_P(value) > YAC_ENTRY_MAX_ORIG_LEN) {
 						php_error_docref(NULL, E_WARNING, "Value is too long(%ld bytes) to be stored", Z_STRLEN_P(value));
-						if (prefix->len) {
-							zend_string_release(prefix_key);
-						}
 						return ret;
 					}
 
@@ -169,27 +193,21 @@ static int yac_add_impl(zend_string *prefix, zend_string *key, zval *value, int
 					if (!compressed_len || compressed_len > Z_STRLEN_P(value)) {
 						php_error_docref(NULL, E_WARNING, "Compression failed");
 						efree(compressed);
-						if (prefix->len) {
-							zend_string_release(prefix_key);
-						}
 						return ret;
 					}
 
 					if (compressed_len > YAC_STORAGE_MAX_ENTRY_LEN) {
 						php_error_docref(NULL, E_WARNING, "Value is too long(%ld bytes) to be stored", Z_STRLEN_P(value));
 						efree(compressed);
-						if (prefix->len) {
-							zend_string_release(prefix_key);
-						}
 						return ret;
 					}
 
 					flag |= YAC_ENTRY_COMPRESSED;
 					flag |= (Z_STRLEN_P(value) << YAC_ENTRY_ORIG_LEN_SHIT);
-					ret = yac_storage_update(ZSTR_VAL(key), ZSTR_LEN(key), compressed, compressed_len, flag, ttl, add, tv);
+					ret = yac_storage_update(key, key_len, compressed, compressed_len, flag, ttl, add, tv);
 					efree(compressed);
 				} else {
-					ret = yac_storage_update(ZSTR_VAL(key), ZSTR_LEN(key), Z_STRVAL_P(value), Z_STRLEN_P(value), flag, ttl, add, tv);
+					ret = yac_storage_update(key, key_len, Z_STRVAL_P(value), Z_STRLEN_P(value), flag, ttl, add, tv);
 				}
 			}
 			break;
@@ -200,22 +218,14 @@ static int yac_add_impl(zend_string *prefix, zend_string *key, zval *value, int
 		case IS_OBJECT:
 			{
 				smart_str buf = {0};
-#if ENABLE_MSGPACK
-				if (yac_serializer_msgpack_pack(value, &buf, &msg))
-#else
 
-				if (yac_serializer_php_pack(value, &buf, &msg))
-#endif
-				{
+				if (yac_serializer(value, &buf, &msg)) {
 					if (buf.s->len > YAC_G(compress_threshold) || buf.s->len > YAC_STORAGE_MAX_ENTRY_LEN) {
 						int compressed_len;
 						char *compressed;
 
 						if (buf.s->len > YAC_ENTRY_MAX_ORIG_LEN) {
 							php_error_docref(NULL, E_WARNING, "Value is too big to be stored");
-							if (prefix->len) {
-								zend_string_release(prefix_key);
-							}
 							return ret;
 						}
 
@@ -224,27 +234,21 @@ static int yac_add_impl(zend_string *prefix, zend_string *key, zval *value, int
 						if (!compressed_len || compressed_len > buf.s->len) {
 							php_error_docref(NULL, E_WARNING, "Compression failed");
 							efree(compressed);
-							if (prefix->len) {
-								zend_string_release(prefix_key);
-							}
 							return ret;
 						}
 
 						if (compressed_len > YAC_STORAGE_MAX_ENTRY_LEN) {
 							php_error_docref(NULL, E_WARNING, "Value is too big to be stored");
 							efree(compressed);
-							if (prefix->len) {
-								zend_string_release(prefix_key);
-							}
 							return ret;
 						}
 
 						flag |= YAC_ENTRY_COMPRESSED;
 						flag |= (buf.s->len << YAC_ENTRY_ORIG_LEN_SHIT);
-						ret = yac_storage_update(ZSTR_VAL(key), ZSTR_LEN(key), compressed, compressed_len, flag, ttl, add, tv);
+						ret = yac_storage_update(key, key_len, compressed, compressed_len, flag, ttl, add, tv);
 						efree(compressed);
 					} else {
-						ret = yac_storage_update(ZSTR_VAL(key), ZSTR_LEN(key), ZSTR_VAL(buf.s), ZSTR_LEN(buf.s), flag, ttl, add, tv);
+						ret = yac_storage_update(key, key_len, ZSTR_VAL(buf.s), ZSTR_LEN(buf.s), flag, ttl, add, tv);
 					}
 					smart_str_free(&buf);
 				} else {
@@ -261,15 +265,11 @@ static int yac_add_impl(zend_string *prefix, zend_string *key, zval *value, int
 			break;
 	}
 
-	if (prefix->len) {
-		zend_string_release(prefix_key);
-	}
-
 	return ret;
 }
 /* }}} */
 
-static int yac_add_multi_impl(zend_string *prefix, zval *kvs, int ttl, int add) /* {{{ */ {
+static int yac_add_multi_impl(yac_object *yac, zval *kvs, int ttl, int add) /* {{{ */ {
 	HashTable *ht = Z_ARRVAL_P(kvs);
 	zend_string *key;
 	zend_ulong idx;
@@ -281,7 +281,7 @@ static int yac_add_multi_impl(zend_string *prefix, zval *kvs, int ttl, int add)
 			key = strpprintf(0, "%lu", idx);
 			should_free = 1;
 		}
-		if (yac_add_impl(prefix, key, value, ttl, add)) {
+		if (yac_add_impl(yac, key, value, ttl, add)) {
 			if (should_free) {
 				zend_string_release(key);
 			}
@@ -298,25 +298,19 @@ static int yac_add_multi_impl(zend_string *prefix, zval *kvs, int ttl, int add)
 }
 /* }}} */
 
-static zval * yac_get_impl(zend_string *prefix, zend_string *key, uint32_t *cas, zval *rv) /* {{{ */ {
+static zval* yac_get_impl(yac_object *yac, zend_string *name, uint32_t *cas, zval *rv) /* {{{ */ {
 	uint32_t flag, size = 0;
 	char *data, *msg;
 	time_t tv;
-	zend_string *prefix_key;
+	const char *key;
+	size_t key_len;
 
-	if ((ZSTR_LEN(key) + prefix->len) > YAC_STORAGE_MAX_KEY_LEN) {
-		php_error_docref(NULL, E_WARNING, "Key%s can not be longer than %d bytes",
-				prefix->len? "(include prefix)" : "", YAC_STORAGE_MAX_KEY_LEN);
+	if ((key = yac_assemble_key(yac, name, &key_len)) == NULL) {
 		return NULL;
 	}
 
-	if (prefix->len) {
-		prefix_key = strpprintf(YAC_STORAGE_MAX_KEY_LEN, "%s%s", ZSTR_VAL(prefix), ZSTR_VAL(key));	
-		key = prefix_key;
-	}
-
 	tv = time(NULL);
-	if (yac_storage_find(ZSTR_VAL(key), ZSTR_LEN(key), &data, &size, &flag, (int *)cas, tv)) {
+	if (yac_storage_find(key, key_len, &data, &size, &flag, (int *)cas, tv)) {
 		switch ((flag & YAC_ENTRY_TYPE_MASK)) {
 			case IS_NULL:
 				if (size == sizeof(int)) {
@@ -362,9 +356,6 @@ static zval * yac_get_impl(zend_string *prefix, zend_string *key, uint32_t *cas,
 							php_error_docref(NULL, E_WARNING, "Decompression failed");
 							efree(data);
 							efree(origin);
-							if (prefix->len) {
-								zend_string_release(prefix_key);
-							}
 							return NULL;
 						}
 						ZVAL_STRINGL(rv, origin, length);
@@ -390,23 +381,13 @@ static zval * yac_get_impl(zend_string *prefix, zend_string *key, uint32_t *cas,
 							php_error_docref(NULL, E_WARNING, "Decompression failed");
 							efree(data);
 							efree(origin);
-							if (prefix->len) {
-								zend_string_release(prefix_key);
-							}
 							return NULL;
 						}
 						efree(data);
 						data = origin;
 						size = length;
 					}
-#if ENABLE_MSGPACK
-					rv = yac_serializer_msgpack_unpack(data, size, &msg, rv);
-#else
-					rv = yac_serializer_php_unpack(data, size, &msg, rv);
-#endif
-					if (!rv) {
-						php_error_docref(NULL, E_WARNING, "Unserialization failed");
-					}
+					rv = yac_unserializer(data, size, &msg, rv);
 					efree(data);
 				}
 				break;
@@ -419,15 +400,11 @@ static zval * yac_get_impl(zend_string *prefix, zend_string *key, uint32_t *cas,
 		rv = NULL;
 	}
 
-	if (prefix->len) {
-		zend_string_release(prefix_key);
-	}
-
 	return rv;
 }
 /* }}} */
 
-static zval * yac_get_multi_impl(zend_string *prefix, zval *keys, zval *cas, zval *rv) /* {{{ */ {
+static zval* yac_get_multi_impl(yac_object *yac, zval *keys, zval *cas, zval *rv) /* {{{ */ {
 	zval *value;
 	HashTable *ht = Z_ARRVAL_P(keys);
 
@@ -435,27 +412,25 @@ static zval * yac_get_multi_impl(zend_string *prefix, zval *keys, zval *cas, zva
 
 	ZEND_HASH_FOREACH_VAL(ht, value) {
 		uint32_t lcas = 0;
-		zval *v, tmp_rv;
-
-		ZVAL_UNDEF(&tmp_rv);
+		zval *v, tmp;
 
 		switch (Z_TYPE_P(value)) {
 			case IS_STRING:
-				if ((v = yac_get_impl(prefix, Z_STR_P(value), &lcas, &tmp_rv)) && !Z_ISUNDEF(tmp_rv)) {
+				if ((v = yac_get_impl(yac, Z_STR_P(value), &lcas, &tmp))) {
 					zend_symtable_update(Z_ARRVAL_P(rv), Z_STR_P(value), v);
 				} else {
-					ZVAL_FALSE(&tmp_rv);
-					zend_symtable_update(Z_ARRVAL_P(rv), Z_STR_P(value), &tmp_rv);
+					ZVAL_FALSE(&tmp);
+					zend_symtable_update(Z_ARRVAL_P(rv), Z_STR_P(value), &tmp);
 				}
 				continue;
 			default:
 				{
 					zend_string *key = zval_get_string(value);
-					if ((v = yac_get_impl(prefix, key, &lcas, &tmp_rv)) && !Z_ISUNDEF(tmp_rv)) {
+					if ((v = yac_get_impl(yac, key, &lcas, &tmp))) {
 						zend_symtable_update(Z_ARRVAL_P(rv), key, v);
 					} else {
-						ZVAL_FALSE(&tmp_rv);
-						zend_symtable_update(Z_ARRVAL_P(rv), key, &tmp_rv);
+						ZVAL_FALSE(&tmp);
+						zend_symtable_update(Z_ARRVAL_P(rv), key, &tmp);
 					}
 					zend_string_release(key);
 				}
@@ -467,48 +442,96 @@ static zval * yac_get_multi_impl(zend_string *prefix, zval *keys, zval *cas, zva
 }
 /* }}} */
 
-void yac_delete_impl(char *prefix, uint32_t prefix_len, char *key, uint32_t len, int ttl) /* {{{ */ {
-	char buf[YAC_STORAGE_MAX_KEY_LEN];
+static int yac_delete_impl(yac_object *yac, zend_string *name, int ttl) /* {{{ */ {
 	time_t tv = 0;
+	const char *key;
+	size_t key_len;
 
-	if ((len + prefix_len) > YAC_STORAGE_MAX_KEY_LEN) {
-		php_error_docref(NULL, E_WARNING, "Key%s can not be longer than %d bytes",
-				prefix_len? "(include prefix)" : "", YAC_STORAGE_MAX_KEY_LEN);
-		return;
-	}
-
-	if (prefix_len) {
-		len = snprintf(buf, sizeof(buf), "%s%s", prefix, key);
-		key = (char *)buf;
+	if ((key = yac_assemble_key(yac, name, &key_len)) == NULL) {
+		return 0;
 	}
 
 	if (ttl) {
 		tv = (zend_ulong)time(NULL);
 	}
 
-	yac_storage_delete(key, len, ttl, tv);
+	return yac_storage_delete(key, key_len, ttl, tv);
 }
 /* }}} */
 
-static void yac_delete_multi_impl(char *prefix, uint32_t prefix_len, zval *keys, int ttl) /* {{{ */ {
+static int yac_delete_multi_impl(yac_object *yac, zval *keys, int ttl) /* {{{ */ {
 	HashTable *ht = Z_ARRVAL_P(keys);
+	int ret = 1;
 	zval *value;
 
 	ZEND_HASH_FOREACH_VAL(ht, value) {
 		switch (Z_TYPE_P(value)) {
 			case IS_STRING:
-				yac_delete_impl(prefix, prefix_len, Z_STRVAL_P(value), Z_STRLEN_P(value), ttl);
+			    ret = ret & yac_delete_impl(yac, Z_STR_P(value), ttl);
 				continue;
 			default:
 				{
-					zval copy;
-					zend_make_printable_zval(value, &copy);
-					yac_delete_impl(prefix, prefix_len, Z_STRVAL(copy), Z_STRLEN(copy), ttl);
-					zval_dtor(&copy);
+					zend_string *key = zval_get_string(value);
+					ret = ret & yac_delete_impl(yac, key, ttl);
+					zend_string_release(key);
 				}
 				continue;
 		}
 	} ZEND_HASH_FOREACH_END();
+
+	return ret;
+}
+/* }}} */
+
+static zend_object *yac_object_new(zend_class_entry *ce) /* {{{ */ {
+	yac_object *yac = emalloc(sizeof(yac_object) + zend_object_properties_size(ce));
+
+	if (!YAC_G(enable)) {
+		zend_throw_exception(NULL, "Yac is not enabled", 0);
+	}
+
+	zend_object_std_init(&yac->std, ce);
+	yac->std.handlers = &yac_obj_handlers;
+	yac->prefix_len = 0;
+
+
+	return &yac->std;
+}
+/* }}} */
+
+static void yac_object_free(zend_object *object) /* {{{ */ {
+	zend_object_std_dtor(object);
+}
+/* }}} */
+
+static zval* yac_read_property_ptr(zval *zobj, zval *name, int type, void **cache_slot) /* {{{ */ {
+	return &EG(error_zval);
+}
+/* }}} */
+
+static zval* yac_read_property(zval *zobj, zval *name, int type, void **cache_slot, zval *rv) /* {{{ */ {
+	if (UNEXPECTED(type == BP_VAR_RW||type == BP_VAR_W)) {
+		return &EG(error_zval);
+	}
+
+	if (yac_get_impl(Z_YACOBJ_P(zobj), Z_STR_P(name), NULL, rv)) {
+		return rv;
+	}
+
+	return &EG(uninitialized_zval);
+}
+/* }}} */
+
+static YAC_WHANDLER yac_write_property(zval *zobj, zval *name, zval *value, void **cache_slot) /* {{{ */ {
+	yac_add_impl(Z_YACOBJ_P(zobj), Z_STR_P(name), value, 0, 0);
+    Z_TRY_ADDREF_P(value);
+
+	YAC_WHANDLER_RET(value);
+}
+/* }}} */
+
+static void yac_unset_property(zval *zobj, zval *name, void **cache_slot) /* {{{ */ {
+	yac_delete_impl(Z_YACOBJ_P(zobj), Z_STR_P(name), 0);
 }
 /* }}} */
 
@@ -517,33 +540,30 @@ static void yac_delete_multi_impl(char *prefix, uint32_t prefix_len, zval *keys,
 PHP_METHOD(yac, __construct) {
 	zend_string *prefix = NULL;
 
-	if (!YAC_G(enable)) {
-		return;
-	}
-
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|S", &prefix) == FAILURE) {
 		return;
 	}
 
-	if (!prefix) {
-		return;
+	if (prefix && ZSTR_LEN(prefix)) {
+		yac_object *yac;
+		if (ZSTR_LEN(prefix) > YAC_STORAGE_MAX_KEY_LEN) {
+			zend_throw_exception_ex(NULL, 0,
+					"Prefix '%s' exceed max key length '%d' bytes", ZSTR_VAL(prefix), YAC_STORAGE_MAX_KEY_LEN);
+			return;
+		}
+		yac = Z_YACOBJ_P(getThis());
+		yac->prefix_len = ZSTR_LEN(prefix);
+		memcpy(yac->prefix, ZSTR_VAL(prefix), ZSTR_LEN(prefix));
 	}
-
-	zend_update_property_str(yac_class_ce, getThis(), ZEND_STRL(YAC_CLASS_PROPERTY_PREFIX), prefix);
-
 }
 /* }}} */
 
 /** {{{ proto public Yac::add(mixed $keys, mixed $value[, int $ttl])
 */
 PHP_METHOD(yac, add) {
-	long ttl = 0;
-	zval rv, *keys, *prefix, *value = NULL;
-	uint32_t ret;
-
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
+	zend_long ttl = 0;
+	zval *keys, *value = NULL;
+	int ret;
 
 	switch (ZEND_NUM_ARGS()) {
 		case 1:
@@ -574,17 +594,14 @@ PHP_METHOD(yac, add) {
 			WRONG_PARAM_COUNT;
 	}
 
-	prefix = zend_read_property(yac_class_ce, getThis(), ZEND_STRL(YAC_CLASS_PROPERTY_PREFIX), 0, &rv);
-
 	if (Z_TYPE_P(keys) == IS_ARRAY) {
-		ret = yac_add_multi_impl(Z_STR_P(prefix), keys, ttl, 1);
+		ret = yac_add_multi_impl(Z_YACOBJ_P(getThis()), keys, ttl, 1);
 	} else if (Z_TYPE_P(keys) == IS_STRING) {
-		ret = yac_add_impl(Z_STR_P(prefix), Z_STR_P(keys), value, ttl, 1);
+		ret = yac_add_impl(Z_YACOBJ_P(getThis()), Z_STR_P(keys), value, ttl, 1);
 	} else {
-		zval copy;
-		zend_make_printable_zval(keys, &copy);
-		ret = yac_add_impl(Z_STR_P(prefix), Z_STR(copy), value, ttl, 1);
-		zval_dtor(&copy);
+		zend_string *key = zval_get_string(keys);
+		ret = yac_add_impl(Z_YACOBJ_P(getThis()), key, value, ttl, 1);
+		zend_string_release(key);
 	}
 
 	RETURN_BOOL(ret);
@@ -594,13 +611,9 @@ PHP_METHOD(yac, add) {
 /** {{{ proto public Yac::set(mixed $keys, mixed $value[, int $ttl])
 */
 PHP_METHOD(yac, set) {
-    long ttl = 0;
-	zval rv, *keys, *prefix, *value = NULL;
-	uint32_t ret;
-
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
+    zend_long ttl = 0;
+	zval *keys, *value = NULL;
+	int ret;
 
 	switch (ZEND_NUM_ARGS()) {
 		case 1:
@@ -631,68 +644,38 @@ PHP_METHOD(yac, set) {
 			WRONG_PARAM_COUNT;
 	}
 
-	prefix = zend_read_property(yac_class_ce, getThis(), ZEND_STRL(YAC_CLASS_PROPERTY_PREFIX), 0, &rv);
-
 	if (Z_TYPE_P(keys) == IS_ARRAY) {
-		ret = yac_add_multi_impl(Z_STR_P(prefix), keys, ttl, 0);
+		ret = yac_add_multi_impl(Z_YACOBJ_P(getThis()), keys, ttl, 0);
 	} else if (Z_TYPE_P(keys) == IS_STRING) {
-		ret = yac_add_impl(Z_STR_P(prefix), Z_STR_P(keys), value, ttl, 0);
+		ret = yac_add_impl(Z_YACOBJ_P(getThis()), Z_STR_P(keys), value, ttl, 0);
 	} else {
-		zval copy;
-		zend_make_printable_zval(keys, &copy);
-		ret = yac_add_impl(Z_STR_P(prefix), Z_STR(copy), value, ttl, 0);
-		zval_dtor(&copy);
+		zend_string *key = zval_get_string(keys);
+		ret = yac_add_impl(Z_YACOBJ_P(getThis()), key, value, ttl, 0);
+		zend_string_release(key);
 	}
 
 	RETURN_BOOL(ret);
 }
 /* }}} */
 
-/** {{{ proto public Yac::__set(string $name, mixed $value)
-*/
-PHP_METHOD(yac, __set) {
-	zend_string *key;
-	zval *prefix, *value, rv;
-
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
-
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "Sz", &key, &value) == FAILURE) {
-		return;
-	}
-
-	prefix = zend_read_property(yac_class_ce, getThis(), ZEND_STRL(YAC_CLASS_PROPERTY_PREFIX), 0, &rv);
-
-	yac_add_impl(Z_STR_P(prefix), key, value, 0, 0);
-}
-/* }}} */
-
 /** {{{ proto public Yac::get(mixed $keys[, int &$cas])
 */
 PHP_METHOD(yac, get) {
 	uint32_t lcas = 0;
-	zval rv, *ret, *keys, *prefix, *cas = NULL;
-
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
+	zval *ret, *keys, *cas = NULL;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z|z", &keys, &cas) == FAILURE) {
 		return;
 	}
 
-	prefix = zend_read_property(yac_class_ce, getThis(), ZEND_STRL(YAC_CLASS_PROPERTY_PREFIX), 0, &rv);
-
 	if (Z_TYPE_P(keys) == IS_ARRAY) {
-		ret = yac_get_multi_impl(Z_STR_P(prefix), keys, cas, return_value);
+		ret = yac_get_multi_impl(Z_YACOBJ_P(getThis()), keys, cas, return_value);
 	} else if (Z_TYPE_P(keys) == IS_STRING) {
-		ret = yac_get_impl(Z_STR_P(prefix), Z_STR_P(keys), &lcas, return_value);
+		ret = yac_get_impl(Z_YACOBJ_P(getThis()), Z_STR_P(keys), &lcas, return_value);
 	} else {
-		zval copy;
-		zend_make_printable_zval(keys, &copy);
-		ret = yac_get_impl(Z_STR_P(prefix), Z_STR(copy), &lcas, return_value);
-		zval_dtor(&copy);
+		zend_string *key = zval_get_string(keys);
+		ret = yac_get_impl(Z_YACOBJ_P(getThis()), key, &lcas, return_value);
+		zend_string_release(key);
 	}
 
 	if (ret == NULL) {
@@ -701,61 +684,28 @@ PHP_METHOD(yac, get) {
 }
 /* }}} */
 
-/** {{{ proto public Yac::__get(string $name)
-*/
-PHP_METHOD(yac, __get) {
-	zval *prefix, rv;
-	zend_string *key;
-	uint32_t lcas = 0;
-
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
-
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S", &key) == FAILURE) {
-		return;
-	}
-
-	prefix = zend_read_property(yac_class_ce, getThis(), ZEND_STRL(YAC_CLASS_PROPERTY_PREFIX), 0, &rv);
-
-	if (yac_get_impl(Z_STR_P(prefix), key, &lcas, return_value) == NULL) {
-		RETURN_FALSE;
-	}
-}
-/* }}} */
-
 /** {{{ proto public Yac::delete(mixed $key[, int $delay = 0])
 */
 PHP_METHOD(yac, delete) {
-	long time = 0;
-	zval *keys, *prefix, rv;
-	char *sprefix = NULL;
-	uint32_t prefix_len = 0;
-
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
+	zend_long time = 0;
+	zval *keys;
+	int ret;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z|l", &keys, &time) == FAILURE) {
 		return;
 	}
 
-	prefix = zend_read_property(yac_class_ce, getThis(), ZEND_STRL(YAC_CLASS_PROPERTY_PREFIX), 0, &rv);
-	sprefix = Z_STRVAL_P(prefix);
-	prefix_len = Z_STRLEN_P(prefix);
-
 	if (Z_TYPE_P(keys) == IS_ARRAY) {
-		yac_delete_multi_impl(sprefix, prefix_len, keys, time);
+		ret = yac_delete_multi_impl(Z_YACOBJ_P(getThis()), keys, time);
 	} else if (Z_TYPE_P(keys) == IS_STRING) {
-		yac_delete_impl(sprefix, prefix_len, Z_STRVAL_P(keys), Z_STRLEN_P(keys), time);
+		ret = yac_delete_impl(Z_YACOBJ_P(getThis()), Z_STR_P(keys), time);
 	} else {
-		zval copy;
-		zend_make_printable_zval(keys, &copy);
-		yac_delete_impl(sprefix, prefix_len, Z_STRVAL(copy), Z_STRLEN(copy), time);
-		zval_dtor(&copy);
+		zend_string *key = zval_get_string(keys);
+		ret = yac_delete_impl(Z_YACOBJ_P(getThis()), key, time);
+		zend_string_release(key);
 	}
 
-	RETURN_TRUE;
+	RETURN_BOOL(ret);
 }
 /* }}} */
 
@@ -763,10 +713,6 @@ PHP_METHOD(yac, delete) {
 */
 PHP_METHOD(yac, flush) {
 
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
-
 	yac_storage_flush();
 
 	RETURN_TRUE;
@@ -778,10 +724,6 @@ PHP_METHOD(yac, flush) {
 PHP_METHOD(yac, info) {
 	yac_storage_info *inf;
 
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
-	
 	inf = yac_storage_get_info();
 
 	array_init(return_value);
@@ -800,6 +742,7 @@ PHP_METHOD(yac, info) {
 	add_assoc_long(return_value, "slots_used", inf->slots_num);
 
 	yac_storage_free_info(inf);
+	return;
 }
 /* }}} */
 
@@ -809,10 +752,6 @@ PHP_METHOD(yac, dump) {
 	long limit = 100;
 	yac_item_list *list, *l;
 
-	if (!YAC_G(enable)) {
-		RETURN_FALSE;
-	}
-	
 	array_init(return_value);
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|l", &limit) == FAILURE) {
@@ -908,9 +847,7 @@ zend_function_entry yac_methods[] = {
 	PHP_ME(yac, __construct, arginfo_yac_constructor, ZEND_ACC_PUBLIC|ZEND_ACC_CTOR)
 	PHP_ME(yac, add, arginfo_yac_add, ZEND_ACC_PUBLIC)
 	PHP_ME(yac, set, arginfo_yac_add, ZEND_ACC_PUBLIC)
-	PHP_ME(yac, __set, arginfo_yac_setter, ZEND_ACC_PUBLIC)
 	PHP_ME(yac, get, arginfo_yac_get, ZEND_ACC_PUBLIC)
-	PHP_ME(yac, __get, arginfo_yac_getter, ZEND_ACC_PUBLIC)
 	PHP_ME(yac, delete, arginfo_yac_delete, ZEND_ACC_PUBLIC)
 	PHP_ME(yac, flush, arginfo_yac_void, ZEND_ACC_PUBLIC)
 	PHP_ME(yac, info, arginfo_yac_void, ZEND_ACC_PUBLIC)
@@ -944,7 +881,7 @@ PHP_MINIT_FUNCTION(yac)
 
 	REGISTER_INI_ENTRIES();
 
-	if(!YAC_G(enable_cli) && !strcmp(sapi_module.name, "cli")) {
+	if (!YAC_G(enable_cli) && !strcmp(sapi_module.name, "cli")) {
 		YAC_G(enable) = 0;
 	}
 
@@ -959,15 +896,58 @@ PHP_MINIT_FUNCTION(yac)
 	REGISTER_LONG_CONSTANT("YAC_MAX_KEY_LEN", YAC_STORAGE_MAX_KEY_LEN, CONST_PERSISTENT | CONST_CS);
 	REGISTER_LONG_CONSTANT("YAC_MAX_VALUE_RAW_LEN", YAC_ENTRY_MAX_ORIG_LEN, CONST_PERSISTENT | CONST_CS);
 	REGISTER_LONG_CONSTANT("YAC_MAX_RAW_COMPRESSED_LEN", YAC_STORAGE_MAX_ENTRY_LEN, CONST_PERSISTENT | CONST_CS);
-#if ENABLE_MSGPACK
-	REGISTER_STRINGL_CONSTANT("YAC_SERIALIZER", "MSGPACK", sizeof("MSGPACK") -1, CONST_PERSISTENT | CONST_CS);
-#else
-	REGISTER_STRINGL_CONSTANT("YAC_SERIALIZER", "PHP", sizeof("PHP") -1, CONST_PERSISTENT | CONST_CS);
+	REGISTER_LONG_CONSTANT("YAC_SERIALIZER_PHP", YAC_SERIALIZER_PHP, CONST_PERSISTENT | CONST_CS);
+#if YAC_ENABLE_MSGPACK
+	REGISTER_LONG_CONSTANT("YAC_SERIALIZER_MSGPACK", YAC_SERIALIZER_MSGPACK, CONST_PERSISTENT | CONST_CS);
 #endif
+#if YAC_ENABLE_IGBINARY
+	REGISTER_LONG_CONSTANT("YAC_SERIALIZER_IGBINARY", YAC_SERIALIZER_IGBINARY, CONST_PERSISTENT | CONST_CS);
+#endif
+#if YAC_ENABLE_JSON
+	REGISTER_LONG_CONSTANT("YAC_SERIALIZER_JSON", YAC_SERIALIZER_JSON, CONST_PERSISTENT | CONST_CS);
+#endif
+
+#if YAC_ENABLE_MSGPACK
+	if (strcmp(YAC_G(serializer), "msgpack") == 0) {
+		yac_serializer = yac_serializer_msgpack_pack;
+		yac_unserializer = yac_serializer_msgpack_unpack;
+		REGISTER_LONG_CONSTANT("YAC_SERIALIZER", YAC_SERIALIZER_MSGPACK, CONST_PERSISTENT | CONST_CS);
+	} else
+#endif
+#if YAC_ENABLE_IGBINARY
+	if (strcmp(YAC_G(serializer), "igbinary") == 0) {
+		yac_serializer = yac_serializer_igbinary_pack;
+		yac_unserializer = yac_serializer_igbinary_unpack;
+		REGISTER_LONG_CONSTANT("YAC_SERIALIZER", YAC_SERIALIZER_IGBINARY, CONST_PERSISTENT | CONST_CS);
+	} else
+#endif
+#if YAC_ENABLE_JSON
+	if (strcmp(YAC_G(serializer), "json") == 0) {
+		yac_serializer = yac_serializer_json_pack;
+		yac_unserializer = yac_serializer_json_unpack;
+		REGISTER_LONG_CONSTANT("YAC_SERIALIZER", YAC_SERIALIZER_JSON, CONST_PERSISTENT | CONST_CS);
+	} else
+#endif
+	{
+		yac_serializer = yac_serializer_php_pack;
+		yac_unserializer = yac_serializer_php_unpack;
+		REGISTER_LONG_CONSTANT("YAC_SERIALIZER", YAC_SERIALIZER_PHP, CONST_PERSISTENT | CONST_CS);
+	}
 
 	INIT_CLASS_ENTRY(ce, "Yac", yac_methods);
 	yac_class_ce = zend_register_internal_class(&ce);
-	zend_declare_property_stringl(yac_class_ce, ZEND_STRS(YAC_CLASS_PROPERTY_PREFIX) - 1, "", 0, ZEND_ACC_PROTECTED);
+	yac_class_ce->ce_flags |= ZEND_ACC_FINAL;
+	yac_class_ce->create_object = yac_object_new;
+
+	memcpy(&yac_obj_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
+	yac_obj_handlers.offset = XtOffsetOf(yac_object, std);
+	yac_obj_handlers.free_obj = yac_object_free;
+	if (YAC_G(enable)) {
+		yac_obj_handlers.read_property  = yac_read_property;
+		yac_obj_handlers.write_property = yac_write_property;
+		yac_obj_handlers.unset_property = yac_unset_property;
+		yac_obj_handlers.get_property_ptr_ptr = yac_read_property_ptr;
+	}
 
 	return SUCCESS;
 }
@@ -989,15 +969,26 @@ PHP_MSHUTDOWN_FUNCTION(yac)
  */
 PHP_MINFO_FUNCTION(yac)
 {
+	smart_str names = {0,};
+
 	php_info_print_table_start();
 	php_info_print_table_header(2, "yac support", "enabled");
 	php_info_print_table_row(2, "Version", PHP_YAC_VERSION);
 	php_info_print_table_row(2, "Shared Memory", yac_storage_shared_memory_name());
-#if ENABLE_MSGPACK
-	php_info_print_table_row(2, "Serializer", "msgpack");
-#else
-	php_info_print_table_row(2, "Serializer", "php");
+
+	smart_str_appends(&names, "php");
+#if YAC_ENABLE_MSGPACK
+	smart_str_appends(&names, ", msgpack");
 #endif
+#if YAC_ENABLE_IGBINARY
+	smart_str_appends(&names, ", igbinary");
+#endif
+#if YAC_ENABLE_JSON
+	smart_str_appends(&names, ", json");
+#endif
+	php_info_print_table_row(2, "Serializer", ZSTR_VAL(names.s));
+	smart_str_free(&names);
+
 	php_info_print_table_end();
 
 	DISPLAY_INI_ENTRIES();
@@ -1017,11 +1008,11 @@ PHP_MINFO_FUNCTION(yac)
 		php_info_print_table_row(2, "Total Shared Memory Usage for values(values_memory_size)", buf);
 		snprintf(buf, sizeof(buf), "%d", inf->segment_size);
 		php_info_print_table_row(2, "Size of Shared Memory Segment(segment_size)", buf);
-		snprintf(buf, sizeof(buf), "%ld", inf->segments_num);
+		snprintf(buf, sizeof(buf), "%d", inf->segments_num);
 		php_info_print_table_row(2, "Number of Segments (segment_num)", buf);
-		snprintf(buf, sizeof(buf), "%ld", inf->slots_size);
+		snprintf(buf, sizeof(buf), "%d", inf->slots_size);
 		php_info_print_table_row(2, "Total Slots Number(slots_size)", buf);
-		snprintf(buf, sizeof(buf), "%ld", inf->slots_num);
+		snprintf(buf, sizeof(buf), "%d", inf->slots_num);
 		php_info_print_table_row(2, "Total Used Slots(slots_num)", buf);
 		php_info_print_table_end();
 
@@ -1034,10 +1025,25 @@ PHP_MINFO_FUNCTION(yac)
 ZEND_GET_MODULE(yac)
 #endif
 
+static zend_module_dep yac_module_deps[] = {
+#if YAC_ENABLE_MSGPACK
+	ZEND_MOD_REQUIRED("msgpack")
+#endif
+#if YAC_ENABLE_IGBINARY
+	ZEND_MOD_REQUIRED("igbinary")
+#endif
+#if YAC_ENABLE_JSON
+	ZEND_MOD_REQUIRED("json")
+#endif
+	{NULL, NULL, NULL, 0}
+};
+
 /* {{{ yac_module_entry
  */
 zend_module_entry yac_module_entry = {
-	STANDARD_MODULE_HEADER,
+	STANDARD_MODULE_HEADER_EX,
+	NULL,
+	yac_module_deps,
 	"yac",
 	NULL, /* yac_functions, */
 	PHP_MINIT(yac),
@@ -1062,3 +1068,4 @@ zend_module_entry yac_module_entry = {
  * vim600: noet sw=4 ts=4 fdm=marker
  * vim<600: noet sw=4 ts=4
  */
+
